from __future__ import absolute_import, division, print_function

import json
import warnings
import threading
import switcherlabs

try:
    import urllib.request
except ImportError:
    pass

try:
    import requests
except ImportError:
    requests = None


def new_http_client(*args, **kwargs):
    if requests:
        impl = RequestsClient
    else:
        impl = UrllibClient
        warnings.warn(
            "Warning: the SwitcherLabs library is falling back to urllib "
            "because requests is not installed. urllib SSL implementation "
            "doesn't verify server certificates. For improved security, we "
            "suggest installing requests."
        )

    return impl(*args, **kwargs)


class HTTPClient(object):
    def __init__(self):
        self._thread_local = threading.local()

    def request(self, method, url, headers, post_data=None):
        raise NotImplementedError(
            "HTTPClient subclasses must implement `request`"
        )

    def close(self):
        raise NotImplementedError(
            "HTTPClient subclasses must implement `close`"
        )


class RequestsClient(HTTPClient):
    name = "requests"

    def __init__(self, **kwargs):
        super(RequestsClient, self).__init__(**kwargs)

    def request(self, method, url, headers):
        if getattr(self._thread_local, "session", None) is None:
            self._thread_local.session = requests.Session()

        url = switcherlabs.api_base + url

        response = self._thread_local.session.request(
            method,
            url=url,
            headers=headers,
            timeout=60,
        )

        return response.json()

    def close(self):
        if getattr(self._thread_local, "session", None) is not None:
            self._thread_local.session.close()


class UrllibClient(HTTPClient):
    name = "urllib.request"

    def __init__(self, **kwargs):
        super(UrllibClient, self).__init__(**kwargs)

    def request(self, method, url, headers):
        url = switcherlabs.api_base + url
        req = urllib.request.Request(url, None, headers)

        response = urllib.request.urlopen(req)

        return json.loads(response.read())

    def close(self):
        pass



from __future__ import absolute_import, division, print_function

import base64
import platform
import time

from switcherlabs import http_client, version

api_base = "https://api.switcherlabs.com"
default_http_client = None


class Client(object):
    def __init__(self,
                 api_key=None,
                 state_refresh_rate=60,
                 identity_refresh_rate=5):
        global default_http_client

        self._state_refresh_rate = state_refresh_rate
        self._identity_refresh_rate = identity_refresh_rate

        if default_http_client:
            self._client = default_http_client
        else:
            default_http_client = http_client.new_http_client()
            self._client = default_http_client

        self._api_key = api_key
        self._headers = self._make_headers()

        self._flags = {}
        self._flags_by_id = {}
        self._overrides = {}
        self._identities = {}
        self._lastRefresh = 0

        self._operations = {
            "==": lambda a, b: a == b,
            "!=": lambda a, b: a != b,
            "<": lambda a, b: a < b,
            "<=": lambda a, b: a <= b,
            ">": lambda a, b: a > b,
            ">=": lambda a, b: a >= b,
        }

        self._refresh_state()

    def _make_headers(self):
        headers = {
            'Content-Type': 'application/json'
        }

        sdk_v = version.VERSION
        py_v = platform.python_version()
        user_agent = "switcherlabs-python/%s python/%s" % (sdk_v, py_v)

        headers["User-Agent"] = user_agent

        if self._api_key is not None:
            basic_auth_string = ":%s" % self._api_key
            base64_bytes = base64.b64encode(basic_auth_string.encode('ascii'))
            auth = "Basic %s" % base64_bytes.decode('ascii')
            headers["Authorization"] = auth

        return headers

    def _refresh_state(self):
        now = time.time()

        if self._lastRefresh + self._state_refresh_rate > now:
            return

        response = self._client.request(
            "GET", "/sdk/initialize", self._headers)

        self._flags = {flag['key']: flag for flag in response['flags']}
        self._flags_by_id = {flag['id']: flag for flag in response['flags']}
        self._overrides = {override['key']: override for override in response['overrides']}
        self._lastRefresh = now

        # Remove stale identities
        self._identities = dict(
            filter(
                lambda elem: elem[1]["fetched_at"] + self._identity_refresh_rate > now,
                self._identities.items()
            )
        )

    def _fetch_identity(self, identifier):
        now = time.time()

        if identifier in self._identities \
            and (self._identities[identifier]["fetched_at"] +
                 self._identity_refresh_rate) > now:
            return self._identities[identifier]

        response = self._client.request(
            "GET", "/sdk/identities/%s" % identifier, self._headers)

        response['fetched_at'] = now
        self._identities[identifier] = response

        return response

    def evaluate_flag(self, key, identifier=None):
        """
        Evaluation order:
            1. Identity override (if identifier is specified)
            2. Environment Override
            3. The first passing dynamic rule
            4. The flag default_value
        """

        if key not in self._flags:
            raise Exception("flag requested does not exist")

        if identifier is not None:
            identity = self._fetch_identity(identifier)
            if key in identity["overrides"]:
                return identity["overrides"][key]

        self._refresh_state()

        if key in self._overrides:
            return self._overrides[key]["value"]

        if len(self._flags[key]["dynamic_rules"]):
            for rule in self._flags[key]["dynamic_rules"]:
                expression_flag_key = self._flags_by_id[rule["expression"]["flag_id"]]["key"]
                flag_value = self.evaluate_flag(expression_flag_key, identifier)
                if self._operations[rule["expression"]["op"]](flag_value, rule["expression"]["value"]):
                    return rule["value"]

        return self._flags[key]["value"]



import os
from codecs import open
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

os.chdir(here)

version_contents = {}
with open(os.path.join(here, "switcherlabs", "version.py"), encoding="utf-8") as f:
    exec(f.read(), version_contents)

readme = None
with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    readme = f.read()

setup(
    name="switcherlabs",
    version=version_contents["VERSION"],
    description="Python SDK for SwitcherLabs",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="SwitcherLabs",
    author_email="support@switcherlabs.com",
    url="https://github.com/switcherlabs/switcherlabs-python",
    license="MIT",
    keywords="switcherslabs api feature-flags",
    packages=find_packages(exclude=["tests", "tests.*"]),
    zip_safe=False,
    install_requires=[
        'requests >= 2.20; python_version >= "3.0"',
        'requests[security] >= 2.20; python_version < "3.0"',
    ],
    python_requires="!=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
)
